import numpy as np

from gammapy.visualization import plot_spectrum_datasets_off_regions

import matplotlib.pyplot as plt

from . import utils

from scipy.stats import chi2
from scipy.stats import norm

import astropy.units as u


def fermi_plotter(name, fermi, subplot = None, **kwargs):
    """
    This returns various plots generated from fermipy.ROIPlotter and fermipy.SEDPlotter

    Args:
        name (str): a plot to show
            Options: ["sqrt_ts", "npred", "ts_hist",
                      "data", "model", "sigma",
                      "excess", "resid", "sed"]
            Default: config.yaml
        fermi (class): vtspy.analysis.FermiAnalysis
        subplot: location of a plot
            Default: None
        **kwargs

    Return:
        AxesSubplot
    """
    from fermipy.plotting import ROIPlotter, SEDPlotter

    output = fermi.output
    roi = fermi.gta.roi
    config = fermi.gta.config

    if subplot is None:
        fig = plt.figure(figsize=(14,6))
        subplot = "111"
        ax = plt.gca()
    else:
        ax = None

    kwargs.setdefault('cmap', config['plotting']['cmap'])

    if name == "sqrt_ts":
        sigma_levels = [3, 5, 7] + list(np.logspace(1, 3, 17))
        plotter = ROIPlotter(output['ts']['sqrt_ts'], roi=roi, **kwargs)
        plotter.plot(
            vmin=0, vmax=5, levels=sigma_levels,
            cb_label='Sqrt(TS) [$\sigma$]',
            interpolation='bicubic', subplot=subplot)
        ax = plt.gca()
        ax.set_title('Sqrt(TS)')

    if name == "npred":
        plotter = ROIPlotter(output['ts']['npred'], roi=roi, **kwargs)
        plotter.plot(
            vmin=0, cb_label='NPred [Counts]', interpolation='bicubic')
        ax = plt.gca()
        ax.set_title('NPred')

    if name == "ts_hist":
        ax = plot_ts_hist(output, subplot)
        ax.set_title('TS histogram')

    if name == "data":
        plotter = ROIPlotter(output['resid']['data'],roi=roi)
        plotter.plot(vmin=50,vmax=400,subplot=subplot,cmap='magma')
        ax = plt.gca()
        ax.set_title('Data')

    if name == "model":
        plotter = ROIPlotter(output['resid']['model'],roi=roi)
        plotter.plot(vmin=50,vmax=400,subplot=subplot,cmap='magma')
        ax = plt.gca()
        ax.set_title('Model')

    if name == "sigma":
        plotter = ROIPlotter(output['resid']['sigma'],roi=roi)
        plotter.plot(vmin=-5,vmax=5,levels=[-5,-3,3,5],subplot=subplot,cmap='RdBu_r')
        ax = plt.gca()
        ax.set_title('Significance')

    if name == "excess":
        plotter = ROIPlotter(output['resid']['excess'],roi=roi)
        plotter.plot(vmin=-100,vmax=100,subplot=subplot,cmap='RdBu_r')
        ax = plt.gca()
        ax.set_title('Excess')

    if name == "resid":
        ax = plot_sigma_hist(output, subplot)
        ax.set_title('Residual histogram')

    if name == "sed":
        kwargs.pop('cmap')
        plot_sed(output, **kwargs)
        ax = plt.gca()

    return ax

def plot_ts_hist(output, subplot=None):
    if subplot is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig = plt.gcf()
        ax = fig.add_subplot(subplot)

    bins = np.linspace(0, 25, 101)

    data = np.nan_to_num(output['ts']['ts'].data.T)
    data[data > 25.0] = 25.0
    data[data < 0.0] = 0.0
    n, bins, patches = ax.hist(data.flatten(), bins, density=True,
                               histtype='stepfilled',
                               facecolor='green', alpha=0.75)

    ax.plot(bins, 0.5 * chi2.pdf(bins, 1.0), color='k',
            label=r"$\chi^2_{1} / 2$")
    ax.set_yscale('log')
    ax.set_ylim(1E-4)
    ax.legend(loc='upper right', frameon=False)

    # labels and such
    ax.set_xlabel('TS')
    ax.set_ylabel('Probability')

    return ax

def plot_sigma_hist(output, subplot=None):
    if subplot is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig = plt.gcf()
        ax = fig.add_subplot(subplot)

    nBins = np.linspace(-6, 6, 121)
    data = np.nan_to_num(output['resid']['sigma'].data)

    # find best fit parameters
    mu, sigma = norm.fit(data.flatten())

    # make and draw the histogram
    data[data > 6.0] = 6.0
    data[data < -6.0] = -6.0

    n, bins, patches = ax.hist(data.flatten(), nBins, density=True,
                               histtype='stepfilled',
                               facecolor='green', alpha=0.75)
    # make and draw best fit line
    y = norm.pdf(bins, mu, sigma)
    ax.plot(bins, y, 'r--', linewidth=2, label="Best-fit")
    y = norm.pdf(bins, 0.0, 1.0)
    ax.plot(bins, y, 'k', linewidth=1, label=r"$\mu$ = 0, $\sigma$ = 1")

    # labels and such
    ax.set_xlabel(r'Significance ($\sigma$)')
    ax.set_ylabel('Probability')
    paramtext = 'Gaussian fit:\n'
    paramtext += '$\\mu=%.2f$\n' % mu
    paramtext += '$\\sigma=%.2f$' % sigma
    ax.text(0.05, 0.95, paramtext, verticalalignment='top',
            horizontalalignment='left', transform=ax.transAxes)

    ax.legend()

    return ax


def plot_ROI(veritas=None, fermi=None):
    plt.figure(figsize=(7, 7))

    if veritas is not None:
        ax = veritas._exclusion_mask.plot()
        veritas._on_region.to_pixel(ax.wcs).plot(ax=ax, edgecolor="red")
        plot_spectrum_datasets_off_regions(ax=ax, datasets=veritas.datasets)

    if fermi is not None:
        if veritas is not None:
            geom = Map.create(npix=(150, 150), binsz=0.05, skydir=fermi.target.skydir, proj="CAR", frame="icrs")
            _, ax, _ = geom.plot()
            ax.add_patch(Patches.Rectangle((0, 0), 150, 150,  color="w"))
        #fermi._src_in_roi(ax)

    plt.show(block=False)

def plot_sed(output, show_model=True, show_band=True, show_flux_points=True, erg=False, units="MeV", color="k", **kwargs):

    sed = output["sed"]
    fermi_model = output["sed"]['model_flux']
    m_engs = 10**fermi_model['log_energies']

    if units == "MeV":
        conv_u = 1
    elif units == "GeV":
        conv_u = 1e-3
    elif units == "TeV":
        conv_u = 1e-6

    if erg:
        e2 = m_engs**2.*utils.MeV2Erg
        conv_f = utils.MeV2Erg
        f_units = "erg/cm$^2$/s"
    else:
        e2 = m_engs**2.*conv_u**2.
        conv_f = conv_u**2.
        f_units = f"{units}/cm$^2$/s"

    ul_ts_threshold = kwargs.pop('ul_ts_threshold', 4)
    m = sed['ts'] < ul_ts_threshold
    x = sed['e_ctr']*conv_u
    y = sed['e2dnde']*conv_f

    yerr = sed['e2dnde_err']*conv_f
    yerr_lo = sed['e2dnde_err_lo']*conv_f
    yerr_hi = sed['e2dnde_err_hi']*conv_f
    yul = sed['e2dnde_ul95']*conv_f
    delo = sed['e_ctr'] - sed['e_min']
    dehi = sed['e_max'] - sed['e_ctr']
    xerr0 = np.vstack((delo[m], dehi[m]))*conv_u
    xerr1 = np.vstack((delo[~m], dehi[~m]))*conv_u

    if show_flux_points:
        plt.errorbar(x[~m], y[~m], xerr=xerr1, label="Fermi-LAT",
                     yerr=(yerr_lo[~m], yerr_hi[~m]), ls="", color=color)
        plt.errorbar(x[m], yul[m], xerr=xerr0,
                     yerr=yul[m] * 0.2, uplims=True, ls="", color=color)

    if show_model:
        plt.plot(m_engs*conv_u, fermi_model['dnde'] * e2, color=color, **kwargs)
        if show_band:
            plt.fill_between(m_engs*conv_u, fermi_model['dnde_lo'] * e2, fermi_model['dnde_hi'] * e2,
            alpha=0.2, color=color)

    plt.xscale("log")
    plt.yscale("log")
    plt.legend(fontsize=13)
    plt.grid(which="major", ls="-")
    plt.grid(which="minor", ls=":")

    plt.xlabel(f"Energy [{units}]", fontsize=13)
    plt.ylabel(f"Energy flux [{f_units}]", fontsize=13)

    fig = plt.gcf()
    return fig



from astropy import units as u

from gammapy.maps import Map

from gammapy.modeling.models import SkyModel

import gammapy.modeling.models as gammapy_model

from astropy.coordinates import Distance

import numpy as np


model_dict = {
    # fermipy to gammapy
    "PowerLaw": gammapy_model.PowerLawSpectralModel,
    "LogParabola": gammapy_model.LogParabolaSpectralModel,
    "PLSuperExpCutoff2":gammapy_model.ExpCutoffPowerLawSpectralModel,

    # gammapy to fermipy
    "PowerLawSpectralModel": "Powerlaw",
    "LogParabolaSpectralModel": "LogParabola",
    "ExpCutoffPowerLawSpectralModel": "PLSuperExpCutoff2"
}

params = {
    "PowerLaw": {"Index": ["index", u.dimensionless_unscaled, u.dimensionless_unscaled, -1],
                 "Prefactor": ["amplitude", 1/u.cm**2/u.s/u.MeV,  1/u.cm**2/u.s/u.TeV, 1],
                 "Scale": ["reference",u.MeV, u.TeV, 1]},
    "LogParabola": {"alpha": ["alpha", u.dimensionless_unscaled, u.dimensionless_unscaled, 1],
                    "beta": ["beta", u.dimensionless_unscaled, u.dimensionless_unscaled, 1],
                    "norm": ["amplitude", 1/u.cm**2/u.s/u.MeV,  1/u.cm**2/u.s/u.TeV, 1],
                    "Eb": ["reference", u.MeV, u.TeV, 1]},
    "PLSuperExpCutoff2": {"Index1": ["index", u.dimensionless_unscaled, u.dimensionless_unscaled, -1],
                    "Index2": ["alpha", u.dimensionless_unscaled, u.dimensionless_unscaled, 1],
                    "Expfactor": ["lambda_", 1/u.MeV, 1/u.TeV, 1],
                    "Prefactor": ["amplitude", 1/u.cm**2/u.s/u.MeV,  1/u.cm**2/u.s/u.TeV, 1],
                    "Scale": ["reference", u.MeV, u.TeV, 1]},
}

not_spectral_shape = ["Prefactor", "norm", "Scale"]

def fermipy2gammapy(like, src, fix_pars=False):

    spectral = model_dict[src['SpectrumType']]()
    for par in src.spectral_pars.keys():
        idx = like.par_index(src.name, par)
        fpar = like.model[idx]

        idx = like.par_index(src.name, par)
        fpar = like.model[idx]
        gpar_setup = params[src['SpectrumType']][par]
        gpar = getattr(spectral, gpar_setup[0])
        val = (fpar.getValue()*fpar.getScale()*gpar_setup[1]*gpar_setup[3]).to(gpar_setup[2])
        gpar.value = val.value
        if fix_pars:
            gpar.frozen = True
        else:
            gpar.frozen = not(fpar.isFree())

    spatial = spatial_model(src)

    source = SkyModel(
        spatial_model=spatial,
        spectral_model=spectral,
        name=src.name,
        )
    return source

def gammapy2fermipy(spectral, src=None):
    from fermipy.roi_model import Source

    if src is None:
        new_model = Source("new",
                {"SpectrumType": model_dict[spectral.tag[0]]})
    else:
        new_model = Source(src.name,
                {"ra": src.radec[0],
                 "dec": src.radec[0],
                 "SpectrumType": model_dict[spectral.tag[0]]})

    params = {}
    for par in new_model.params.keys():
        fpar = new_model.params[par]
        gpar_setup = params[new_model['SpectrumType']][par]
        gpar = getattr(spectral, gpar_setup[0])
        val = (gpar.value*gpar_setup[2]).to(gpar_setup[1])
        params[par] = {"value":val.value, "free": not(val.frozen)}

    return params

def fermi_galactic_diffuse_bkg(config, fmodel):
    diffuse_galactic_fermi = Map.read(config['model']['galdiff'][0])
    spatial_model = gammapy_model.TemplateSpatialModel(diffuse_galactic_fermi, normalize=False)
    spectral_model = gammapy_model.PowerLawNormSpectralModel()
    spectral_model.norm.value = fmodel.params["Prefactor"]["value"]
    diffuse_gald = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model,  name="galdiff")
    return diffuse_gald

def fermi_isotropic_diffuse_bkg(config, fmodel):
    diffuse_iso = gammapy_model.create_fermi_isotropic_diffuse_model(
        filename=config['model']['isodiff'][0],
        interp_kwargs={"fill_value": None})
    diffuse_iso.spectral_model.model2.value = fmodel.params["Normalization"]["value"]
    diffuse_iso._name = "isodiff"
    return diffuse_iso

def default_model(model, **kwargs):
    if (model.lower() == "powerlaw") or (model.lower() == "pl"):
        amplitude = kwargs.pop("amplitude", 1e-12)
        index = kwargs.pop("index", 2.5)
        reference = kwargs.pop("reference", 1)

        spectral_model = gammapy_model.PowerLawSpectralModel(
            amplitude=amplitude * u.Unit("cm-2 s-1 TeV-1"),
            index=index,
            reference=reference * u.TeV,
        )
    elif model.lower() == "logparabola":
        amplitude = kwargs.pop("amplitude", 1e-12)
        alpha = kwargs.pop("alpha", 3)
        beta = kwargs.pop("beta", 2)
        reference = kwargs.pop("reference", 1)

        spectral_model = gammapy_model.LogParabolaSpectralModel(
             amplitude=1e-12 * u.Unit("cm-2 s-1 TeV-1"),
             alpha=3,
             beta=2,
             reference=1 * u.TeV,
        )
    elif model.lower() == "agnpy":
        from .external.agnpy import agnpy_spectral_model

        spectral_model = agnpy_spectral_model()

        z = kwargs.pop("z", 0)
        t_var = kwargs.pop("t_var", 0)

        if (z == 0) or (t_var == 0):
            self._logging.error("The source redshfit and variability are required.")
            raise
        d_L = Distance(z=z).to("cm")

        log10_norm_e = kwargs.pop("log10_norm_e", -5)
        p1 = kwargs.pop("p1", -1)
        p2 = kwargs.pop("p2", 3)

        delta_D = kwargs.pop("delta_D", 10)
        log10_B = kwargs.pop("log10_B", 1)

        log10_gamma_b = kwargs.pop("log10_gamma_b", 4)
        log10_gamma_min = kwargs.pop("log10_gamma_min", np.log10(500))
        log10_gamma_max = kwargs.pop("log10_gamma_max", np.log10(1e6))

        spectral_model.z.quantity = z
        spectral_model.z.frozen = True
        spectral_model.d_L.quantity = d_L
        spectral_model.d_L.frozen = True

        spectral_model.delta_D.quantity = delta_D
        spectral_model.log10_B.quantity = log10_B
        spectral_model.t_var.quantity = t_var
        spectral_model.t_var.frozen = True

        spectral_model.log10_norm_e.quantity = log10_norm_e
        spectral_model.log10_norm_e._is_norm = True
        spectral_model.p1.quantity = p1
        spectral_model.p2.quantity = p2
        spectral_model.log10_gamma_b.quantity = log10_gamma_b
        spectral_model.log10_gamma_min.quantity = log10_gamma_min
        spectral_model.log10_gamma_min.frozen = True
        spectral_model.log10_gamma_max.quantity = log10_gamma_max
        spectral_model.log10_gamma_max.frozen = True

    else:
        spectral_model = None
    return spectral_model

def spatial_model(src):
    if src['SpatialModel'] == "PointSource":
        spatial_model = gammapy_model.PointSpatialModel(
        lon_0="{:.3f} deg".format(src['spatial_pars']['RA']['value']), lat_0="{:.3f} deg".format(src['spatial_pars']['DEC']['value']), frame="icrs"
        )

    elif src['SpatialModel'] == "RadialGaussian":
        spatial_model = gammapy_model.GaussianSpatialModel(
        lon_0="{:.3f} deg".format(src['spatial_pars']['RA']['value']), lat_0="{:.3f} deg".format(src['spatial_pars']['DEC']['value']), sigma= "{:.3f} deg".format(src['spatial_pars']['Sigma']['value']), frame="icrs"
        )
    elif src['SpatialModel'] =="SpatialMap":
        spatial_model_temp = Map.read(src['Spatial_Filename'])
        spatial_model_temp.unit = "sr-1"
        spatial_model =  gammapy_model.TemplateSpatialModel(spatial_model_temp, normalize=False)
    else:
        self._logging.error(f"This type of the spatial model is not yet supported; {src['SpatialModel']}")
        raise

    spatial_model.parameters.freeze_all()
    return spatial_model



import os
import numpy as np
import copy

from astropy.time import Time, TimeDelta
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits

import logging
logging.basicConfig(format=('%(asctime)s %(levelname)-8s: %(message)s'), datefmt='%Y-%m-%d %H:%M:%S', level=20)

from pathlib import Path
SCRIPT_DIR = str(Path(__file__).parent.absolute())

MeV2Erg = 1.60218e-6
TeV2Erg = 1.60218

def logger(verbosity = 1):
    """
    Set a log level:

    * 1: info, warning, error
    * 2: debug, info, warning, error
    * 0: warning, error
    * -1: error

    Args:
        verbosity (int)
            Default: 1
    
    Return:
        astorpy.time: UTC time
    """
    levels_dict = {2: 10,
                   1: 20,
                   0: 30,
                   -1: 40,
                   -2: 50}
                   
    level = levels_dict[verbosity]
    logging.getLogger().setLevel(level)
    return logging

def MET2UTC(met, return_astropy=False):
    """
    Convert Fermi MET (Mission Elapsed Time) to UTC (Coordinated 
    Universal Time).

    Args:
        met (float): MET time in seconds
        return_astropy (bool): return astropy.time
    
    Return:
        str, astropy.time (optional): UTC time
    """

    refMET = Time('2001-01-01', format='isot', scale='utc')

    dt = TimeDelta(met, format='sec')

    if return_astropy:
        return (refMET+dt).iso, refMET+dt
    else:
        return (refMET+dt).iso

def UTC2MET(utc):
    """
    Convert UTC to Fermi MET (mission elapsed time).

    Args:
        utc (astorpy.time): UTC time
    
    Return:
        float: MET
    """
    refMET = Time('2001-01-01', format='isot', scale='utc')
    currentTime = Time(utc, format='isot', scale='utc')
    return float((currentTime-refMET).sec)

def METnow():
	"""
    Return the current MET.

    Return:
        float: Fermi MET
    """
	refMET = Time('2001-01-01', format='isot', scale='utc')
	currentTime = Time.now()
	return float((currentTime-refMET).sec)

def MJD2UTC(mjd, return_astropy=False):
    """
    Convert MJD (Modified Julian Day) to UTC.

    Args:
        mjd (astorpy.time): MJD time
        return_astropy (bool): return astropy.time

    Return:
        str, astropy.time (optional): UTC time
    """

    refMJD = Time(mjd, format='mjd')
    if return_astropy:
        return refMJD.isot, refMJD
    else:
        return refMJD.isot

def UTC2MJD(utc):
    """
    Convert UTC to MJD.

    Args:
        mjd (astorpy.time): MJD time
    
    Return:
        float: MJD
    """
    refUTC= Time(utc, format='isot', scale='utc')
    return float(refUTC.mjd)

def CEL2GAL(ra, dec):
	"""
    Convert CEL (celestial) coordinates to GAL (galactic) coordinates.

    Args:
        ra (float): right ascension in degrees
        dec (float): declination in degrees
    
    Return:
        deg, deg
    """
	c = SkyCoord(ra=float(ra)*u.degree, dec=float(dec)*u.degree, frame='icrs')
	return c.galactic.l.deg, c.galactic.b.deg

def GAL2CEL(l, b):
	"""
    Convert MJD (Modified Julian Day) to UTC.

    Args:
        mjd (astorpy.time): MJD time
    
    Return:
        astropy.time: UTC
    """
	c = SkyCoord(l=float(l)*u.degree, b=float(b)*u.degree, frame='galactic')
	return c.icrs.ra.deg, c.icrs.dec.deg 

def bright_source_list(source, save_npy=True):

    if ".npy" in source:
        bright_sources = np.load(source)
    else:
        bright_sources = []
        with open(source) as f: 
            for line in f.readlines()[27:]:
                info = line.split()
                if len(info) == 5:
                    ra = info[0]
                    dec = info[1]
                    bright = info[3]
                    brightbv = info[4]

                    bright_sources.append([float(ra), float(dec), float(bright), float(brightbv)])
            if save_npy:
                source = source.replace(".dat", ".npy")
                np.save(source, bright_sources)
    return np.asarray(bright_sources)

def time_filter(obs, time, time_format=None):
    """
    Filter the observations by time
    
    Args:
        obs (gammapy.observations): the list of observations
        time (list): the list of times
        time_format (str, optional): passed to astropy.Time
    """

    obs = copy.deepcopy(obs)

    if type(time[0]) == Time:
        time_interval = time
    else:
        tmin, tmax = time[0], time[1]
        if time_format == "MET":
            tmin = MET2UTC(tmin)
            tmax = MET2UTC(tmax)
            time_interval= Time([str(tmin),str(tmax)])
        else:
            time_interval= Time([str(tmin),str(tmax)], format=time_format, scale="utc")
    
    newobs = obs.select_time(time_interval)
    
    return newobs, newobs.ids

def LiMaSiginficance(N_on, N_off, alpha, type=1):
    """
    Calculate the Li&Ma significance
    
    Args:
        N_on (int): the number of events from the ON region
        N_off (int): the number of events from the OFF region
        alpha (float): the relative exposure time
        type (int): method for calculating signficance 
            1: Li&MA, 2: chisq
            Default: 1
    Return
        float: Li&Ma significance

    """

    if type == 1:
        temp = N_on*np.log((1.+alpha)/alpha*(N_on/(N_on+N_off)))+N_off*np.log((1+alpha)*(N_off/(N_on+N_off)))
    
        if np.size(temp) != 1:
            for i, t in enumerate(temp):
                if t > 0:
                    temp[i] = np.sqrt(t)
                else:
                    temp[i] = np.nan
        else:
            if temp >0:
                temp = np.sqrt(temp)
            else:
                temp = np.nan

        significance = np.sign(N_on-alpha*N_off)*np.sqrt(2.)*temp
    else:
        significance = (N_on-alpha*N_off)/np.sqrt(alpha*(N_on+N_off))
    return significance


def define_time_intervals(tmin, tmax, binsz=None, nbins=None):
    """
    Define time intervals by either a bin size or the number of bins
    
    Args:
        tmin (float): minimum time
        tmax (float): maximum time
        binsz (astropy.Quantity, optional): bin size with astropy.units
        nbins (int, optional): the number of bins
    
    Return:
        list: time interval (astropy.Time)
    """
    if binsz is not None:
        _, tmin = MJD2UTC(tmin, return_astropy=True)
        _, tmax = MJD2UTC(tmax, return_astropy=True)
        nbins = int((tmax-tmin)*u.second/binsz.to(u.second))
        nbins +=1
        times = tmin + np.arange(nbins) * binsz
    elif nbins is not None:
        _, tmin = MJD2UTC(tmin, return_astropy=True)
        _, tmax = MJD2UTC(tmax, return_astropy=True)
        binsz = ((tmax-tmin)/nbins).to(u.second)
        nbins +=1
        times = tmin + np.arange(nbins) * binsz

    if len(times) == 1:
        time_intervals = [Time([tmin, tmax])]
    else:
        time_intervals = [Time([tstart, tstop]) for tstart, tstop in zip(times[:-1], times[1:])]

    return time_intervals


import os
import numpy as np
import matplotlib.pyplot as plt
import glob
from ..config import JointConfig
from ..utils import logger
from ..plotting import fermi_plotter

from astropy import units as u
from astropy.coordinates import SkyCoord

from gammapy.data import EventList
from gammapy.datasets import MapDataset, Datasets

from gammapy.irf import PSFMap, EDispMap
from gammapy.maps import Map, MapAxis, WcsGeom

from ..model import *

from gammapy.modeling.models import Models

from gammapy.modeling import Fit

from fermipy.gtanalysis import GTAnalysis

import fermipy.wcs_utils as wcs_utils
import fermipy.utils as fermi_utils

from regions import CircleSkyRegion
from pathlib import Path

def generatePSF(config):
    from GtApp import GtApp
    gtpsf = GtApp('gtpsf')
    workdir = config['fileio']['workdir']
    gtpsf["expcube"] = '{}/ltcube_00.fits'.format(workdir)
    gtpsf["outfile"] = '{}/gtpsf_00.fits'.format(workdir)
    gtpsf["irfs"] = config['gtlike']['irfs']
    gtpsf['evtype'] = config['selection']['evtype']
    gtpsf['ra'] = config['selection']['ra']
    gtpsf['dec'] = config['selection']['dec']
    gtpsf['emin'] = config['selection']['emin']
    gtpsf['emax'] = config['selection']['emax']
    gtpsf['chatter'] = 0
    gtpsf.run()

class FermiAnalysis():
    """
    This is to perform a simple Fermi-LAT analysis and to prepare the joint-fit
    analysis. All fermipy.GTanalysis functions and attributes can be accessed
    with the 'gta' arribute. e.g.,

        fermi = FermiAnalysis()
        fermi.gta.optimize()

    All information about the state of the analysis (See fermipy documentation
    for details) will be saved in the numpy array format (npy).

    Args:
        state_file (str): state filename (npy)
            Default: initial
        config_file (str): config filename (yaml)
            Default: config.yaml
        overwrite (bool): overwrite the state
            Default: False
        remove_weak_srcs (bool): remove sources with TS of nan or 0
            Default: False
        construct_dataset (bool): construct dataset for the gammapy analysis
            Default: False
        verbosity (int)
        **kwargs: passed to fermipy.GTAnalysis module
    """

    def __init__(self, state_file = "initial", config_file='config.yaml', overwrite=False, remove_weak_srcs = False, construct_dataset = False, verbosity=True, **kwargs):

        self._verbosity = verbosity
        config = JointConfig.get_config(config_file=config_file).pop("fermi")

        self._logging = logger(self.verbosity)
        self._logging.info("Initialize the Fermi-LAT analysis.")

        self.gta = GTAnalysis(config, logging={'verbosity' : self.verbosity+1}, **kwargs)
        self._outdir = self.gta.config['fileio']['outdir']

        nbin = kwargs.pop("nbin", 6)
        self._energy_bins = MapAxis.from_bounds(self.gta.config["selection"]["emin"],
                self.gta.config["selection"]["emax"],
                nbin=nbin,
                name="energy",
                interp="log", unit="MeV")

        if overwrite or not(os.path.isfile(f"{self._outdir}/{state_file}.fits")):

            if overwrite:
                self._logging.info("Overwrite the Fermi-LAT setup.")
            else:
                self._logging.info("Initial setup and configuration are not found. Performing the data reduction...")

            self._logging.debug("Generate fermipy files.")
            self.gta.setup(overwrite=overwrite)

            self.gta.config["data"]["ltcube"] = f"{self._outdir}/ltcube_00.fits"

            self._logging.debug("Optimize the ROI.")
            self.gta.optimize()

            if remove_weak_srcs:
                self.remove_weak_srcs()

            self._logging.debug("Generate PSF.")
            generatePSF(self.gta.config)

            if construct_dataset:
                self.construct_dataset()

            self.save_state(state_file, init=True, **kwargs)

            self._logging.info("The initial setup and configuration is saved [state_file = {}].".format(state_file))
        else:
            self._logging.info("The setup and configuration is found [state_file = {}]. Loading the configuration...".format(state_file))

            flag = self.load_state(state_file)

            if flag == -1:
                return

        self._test_model = {'Index' : 2.0, 'SpatialModel' : 'PointSource' }
        self._find_target()
        self._logging.info("Completed (Fermi-LAT initialization).")

        if construct_dataset:
            self.construct_dataset()

    @property
    def target(self):
        """
        Return:
            fermipy.roi_model.Source
        """
        return self._target

    @property
    def target_name(self):
        """
        Return:
            str: target name
        """
        return self._target_name

    @property
    def target_id(self):
        """
        Return:
            int: target id
        """
        return self._target_id

    def print_association(self):
        """
        Print sources within ROI and their associations.
        """

        for i, src in enumerate(self.gta.roi.sources):
            if src.name == "isodiff" or src.name=="galdiff":
                continue

            self._logging.info(str(i)+") "+src.name+":"+str(src.associations[1:]))

    def print_target(self):
        """
        Print the target properties
        """
        self._logging.info(self.gta.roi.sources[self.target_id])

    def print_model(self):
        """
        Print source models within ROI
        """
        return self.gta.print_model(loglevel=40)

    def print_params(self, full_output=False):
        """
        Print parameters of sources within ROI
        """
        if full_output:
            return self.gta.print_params(True, loglevel=40)
        else:
            return self.gta.print_params(False, loglevel=40)


    @property
    def verbosity(self):
        """
        Return:
            int
        """
        return self._verbosity


    def peek_events(self):
        """
        Show event information
        """
        if not(hasattr(self, "_gammapy_events")):
            self._logging.error("Run FermiAnalysis.construct_dataset first.")
            return

        self._gammapy_events.peek()

    def peek_irfs(self):
        """
        Show instrument response function (irf) information
        """
        if not(hasattr(self, "datasets")):
            self._logging.error("Run FermiAnalysis.construct_dataset first.")
            return

        e_true = self.datasets.edisp.edisp_map.geom.axes[1]
        e_reco = MapAxis.from_energy_bounds(
            e_true.edges.min(),
            e_true.edges.max(),
            nbin=len(e_true.center),
            name="energy",
        )

        edisp_kernel = self.datasets.edisp.get_edisp_kernel(energy_axis=e_reco)


        f, ax = plt.subplots(2,2, figsize=(10, 6))
        edisp_kernel.plot_bias(ax = ax[0][0])
        ax[0][0].set_xlabel(f"$E_\\mathrm{{True}}$ [MeV]")

        edisp_kernel.plot_matrix(ax = ax[0][1])
        self.datasets.psf.plot_containment_radius_vs_energy(ax = ax[1][0])
        self.datasets.psf.plot_psf_vs_rad(ax = ax[1][1])
        plt.tight_layout()

    def save_state(self, state_file, init=False):
        """
        Save the state

        Args:
            state_file (str): passed to fermipy.write_roi
            init (bool): check whether this is the initial analysis.
                Default: False
        """

        if (init==False) and (state_file == "initial"):
            self._logging.warning("The 'inital' state is overwritten. This is not recommended.")
            self._logging.warning("The original 'inital' state is archived in the '_orig' folder.")
            os.system(f"mkdir ./{self._outdir}/_orig")
            for file in glob.glob(f"{self._outdir}/*initial*"):
                os.sytem(f"mv {file} {self._outdir}/_orig/")

        self.gta.write_roi(state_file, save_model_map=True)
        self._fermi_state = state_file

    def load_state(self, state_file):
        """
        Load the state

        Args:
            state_file (str): passed to fermipy.write_roi
        """
        try:
            self.gta.load_roi(state_file)
            self._fermi_state = state_file
            if os.path.exists(f"{self._outdir}/{state_file}_output.npy"):
                self.output = np.load(f"{self._outdir}/{state_file}_output.npy", allow_pickle=True).item()
        except:
            self._logging.error("The state file does not exist. Check the name again")
            return -1

    def set_target(self, target):
        """
        Set/change the target

        Args:
            target (str or int): target name or id
        """
        if type(target)==int:
            self._target = self.gta.roi.sources[target]
            self._target_id = target
            self._logging.info(f"The target is set to {self.gta.roi.sources[target].name}")
            return
        elif type(target)==str:
            self._find_target(name=target)
            self._logging.info(f"The target is set to {target}")

    def remove_weak_srcs(self, ts_cut=1, npred_cut=0):

        """
        Remove sources within ROI if they are too weak.
        Args:
            ts_cut (float): remove sources with a TS cut
                Default: 1
            npred_cut (float): remove sources with a npred cut
                Default: 0
        """
        N = 0
        for src in self.gta.roi.sources:
            if (src.name == "isodiff") or (src.name=="galdiff") or (src.name == self.target_name):
                continue
            if np.isnan(src['ts']) or src['ts'] < ts_cut or src['npred'] < npred_cut:
                self.gta.delete_source(src.name)
                N+=1
        self._logging.info(f"{N} sources are deleted.")


    def fit(self, state_file="simple", pre_state=None,
        free_all=False, free_target=True,
        remove_weak_srcs=False, fix_index=False,
        min_ts=5, distance=3.0, optimizer = 'NEWMINUIT',
        return_output=False, **kwargs):
        """
        Perform a simple fitting with various cuts

        Args:
            state_file (str): output state filename (npy)
                Default: simple
            free_all (bool): make the target's all parameters free
                Default: True
            free_all (bool): make all sources parameters free
                Default: False
            remove_weak_srcs (bool): remove sources with TS of nan or 0. This setting
                will trigger another round of fit process after the first run.
                Default: False
            fix_index (bool): fix spectral shapes for sources for TS less than min_ts
                Default: False
            min_ts (int): minimum TS value for fixing a spectral shape
                Default: 5
            distance (float): parameters for sources outside of a certain distance
                from the center are fixed, except for the normalization
                Default: 3.0
            optimizer (str): either MINUIT or NEWMINUIT
                Default: NEWMINUIT
            return_output (bool): return the fitting result (dict)
                Default: False
            pre_state (str, optional): input state filename (npy). If not defined, starting from
                the current state.
                Default: None
            **kwargs: passed to fermipy.GTAnalysis.free_sources function

        Return
            dict: the output of the fitting when return_output is True
        """
        if pre_state is not None:
            self.load_state(pre_state)

        if free_all:
            self.gta.free_sources(free=True, **kwargs)
        else:
            self.gta.free_sources(free=False)

            self.gta.free_sources(free=True, distance=distance,  pars='norm', **kwargs)

            if not(fix_index):
                self.gta.free_sources(free=True, minmax_ts=[min_ts, None], **kwargs)

        if free_target:
            self.gta.free_sources_by_name(self.target_name, free=True, pars=None)


        o = self.gta.fit(optimizer=optimizer, reoptimize=True, min_fit_quality=2, verbosity=False)

        if remove_weak_srcs:
            self.remove_weak_srcs()
            o = self.gta.fit(optimizer=optimizer, reoptimize=True, min_fit_quality=2, verbosity=False)

        if o["fit_success"]:
            self._logging.info("Fit successfully ({}).".format(o["fit_quality"]))
        else:
            self._logging.error("Fit failed.")

        self.save_state(state_file)
        self._logging.info(f"The state is saved as '{state_file}'. You can load the state by vtspy.FermiAnalysis('{state_file}').")

        if return_output:
            return o

    def analysis(self, jobs = ["ts", "resid", "sed"], state_file="analyzed", **kwargs):
        """
        Perform various analyses: TS map, Residual map, and SED.

        Args:
            jobs (str or list): list of jobs, 'ts', 'resid', and/or 'sed'.
                Default: ['ts', 'resid', 'sed']
            state_file (str): output state filename (npy)
                Default: analyzed
            **kwargs: passed to GTanalysis.sed
        """

        try:
            output = np.load(f"{self._outdir}/{state_file}_output.npy", allow_pickle=True).item()
        except:
            output = {}

        model = kwargs.get("model", self._test_model)
        energy_bins = kwargs.get("energy_bins", np.log10(self._energy_bins.edges.value))

        free = self.gta.get_free_param_vector()

        if type(jobs) == str:
            jobs = [jobs]

        if "ts" in jobs:
            o = self._ts_map(model=model)
            output['ts'] = o

        if "resid" in jobs:
            o = self._resid_dist(model=model)
            output['resid'] = o

        if "sed" in jobs:
            outfile=state_file+"_sed.fits"
            o = self._calc_sed(outfile=outfile, **kwargs)
            output['sed'] = o

        self.gta.set_free_param_vector(free)

        self.output = output
        np.save(f"{self._outdir}/{state_file}_output", output)

        self.save_state(state_file)
        self._logging.info(f"The state is saved as '{state_file}'. You can load the state by vtspy.FermiAnalysis('{state_file}').")

    def plot(self, output, state_file="analyzed", **kwargs):
        """
        Show various plots: TS map, Residual map, and SED.

        Args:
            output (str or list): list of plots to show
                Options: ["sqrt_ts", "npred", "ts_hist", "data",
                "model", "sigma", "excess", "resid", "sed"]
            state_file (str): read the output (from FermiAnalysis.analysis)
        """

        if not(hasattr(self, "output")):
            try:
                self._logging.info("Loading the output file...")
                self.output = np.load(f"{self._outdir}/{state_file}_output.npy", allow_pickle=True).item()
            except:
                self._logging.error("Run FermiAnalysis.analysis first.")
                return

        list_of_fig = ["sqrt_ts", "npred", "ts_hist",
                        "data", "model", "sigma", "excess", "resid",
                        "sed"]

        if type(output) == str:
            output = [output]

        for o in output:
            if o not in list_of_fig:
                output.remove(o)

        if len(output) == 1:
            sub = "11"
        elif len(output) <= 3:
            sub = "1"+str(len(output))
            f = plt.figure(figsize=(4*len(output), 4))
        elif len(output) == 4:
            sub = "22"
            f = plt.figure(figsize=(8, 8))
        elif len(output) == 6:
            sub = "23"
            f = plt.figure(figsize=(12, 8))

        for i, o in enumerate(output):
            subplot = int(sub+f"{i+1}")
            ax = fermi_plotter(o, self, subplot=subplot, **kwargs)

        plt.tight_layout()


    def find_sources(self, state_file = "wt_new_srcs", re_fit=True, **kwargs):
        """
        Find sources within the ROI (using GTanalysis.find_sources).

        Args:
            state_file (str): output state filename (npy)
                Default: wt_new_srcs
            re_fit (bool): re fit the ROI with new sources
            **kwargs: passed to fit.

        Return:
            dict: output of GTanalysis.find_sources
        """
        srcs = self.gta.find_sources(model=self._test_model, sqrt_ts_threshold=5.0,
                        min_separation=0.5)

        self._logging.info("{} sources are found. They will be added into the model list.".format(len(srcs["sources"])))

        if re_fit:
            self.fit(state_file=state_file, **kwargs)
        else:
            self.save_state(state_file)

        return srcs


    def construct_dataset(self,
                        fix_other_srcs = False,
                        eventlist = "ft1_00.fits",
                        exposure = "bexpmap_00.fits",
                        psf = "gtpsf_00.fits"):

        """
        Construct a dataset for the gammapy analysis.

        Args:
            fix_other_srcs (bool): fix source parameters except for the target
                Default: False
            eventlist (str): event list file (gtapp.maketime)
                Default: ft1_00.fits
            exposure (str): exposure map file (gtapp.gtexpmap)
            psf (str): psf file (gtapp.gtpsf)
        """

        self._logging.info("Loading the Fermi-LAT events...")
        counts = self._load_fermi_events(eventlist=eventlist)
        self._logging.info("Loading the Fermi-LAT IRFs...")
        irf = self._load_fermi_irfs(counts, exposure=exposure, psf=psf)
        self._logging.info("Loading the Fermi-LAT models...")
        models =  self._convert_model(fix_other_srcs=fix_other_srcs)
        self.datasets = MapDataset(
            name="fermi", models=models, counts=counts, exposure=irf["exposure"], psf=irf["psf"], edisp=irf["edisp"]
        )
        self._logging.info("Ready to perform a gammapy analysis.")

    def _find_target(self, name=None):
        if name is None:
            name = self.gta.config['selection']['target']

        flag = False
        for i, src in enumerate(self.gta.roi.sources):
            if src.name == "isodiff" or src.name=="galdiff":
                continue

            for n in src.associations:
                if (n.replace(" ", "") == name) or (n == name):
                    self._target = self.gta.roi.sources[i]
                    self._target_name = self.gta.roi.sources[i].name
                    self._target_id = i
                    list_of_association = src.associations
                    flag = True
                    self.gta.config["selection"]["target"] = self.target_name


            if flag:
                break

        if flag:
            self._logging.info("The target, {}, is associated with {} source(s).".format(self.target_name, len(list_of_association)-1))
            self._logging.debug(list_of_association)
        else:
            self._logging.warning("The target name defined in the config file is not found.")
            self._target = self.gta.roi.sources[0]
            self._target_name = self.target.name
            self._target_id = 0

    def _ts_map(self, model):
        self._logging.info("Generating a TS map...")
        self.gta.free_sources(free=False)
        self.gta.free_sources(pars="norm")
        o = self.gta.tsmap('ts', model=model, write_fits=True, write_npy=True, make_plots=True)
        self._logging.info("Generating the TS map is completed.")
        return o

    def _resid_dist(self, model):
        self._logging.info("Generating a residual distribution...")
        self.gta.free_sources(free=False)
        self.gta.free_sources(pars="norm")
        o = self.gta.residmap('resid',model=model, write_fits=True, write_npy=True)
        self._logging.info("Generating the residual distribution is completed.")
        return o

    def _calc_sed(self, target=None, **kwargs):
        self._logging.info("Generating a SED... ")

        if target is None:
            target = self.target_name

        distance = kwargs.pop("distance", 3.0)

        loge_bins = kwargs.pop("loge_bins",  np.log10(self._energy_bins.edges.value))
        
        outfile = kwargs.pop("outfile", 'sed.fits')

        self.gta.free_sources(free=False)
        self.gta.free_sources(skydir=self.gta.roi[target].skydir, distance=[distance], free=True)
        o = self.gta.sed(self.target.name, outfile=outfile, bin_index=2.2, loge_bins=loge_bins, write_fits=True, write_npy=True, **kwargs)
        self._logging.info("Generating the SED is completed.")
        return o

    def _lightcurve(self, target=None, **kwargs):

        self._logging.info("Generating a light curve...")

        if target is None:
            target = self.target_name

        free_radius = kwargs.pop("free_radius", 3.0)

        self.gta.free_sources(free=False)
        self.gta.free_sources(pars="norm")

        o = self.gta.lightcurve(target, free_radius=free_radius, multithread=True,
            nthread=4, use_scaled_srcmap=True, **kwargs)

        self._logging.info("Generating the lightcurve is completed.")
        return o

    def _load_fermi_events(self, eventlist = "ft1_00.fits"):

        self._gammapy_events = EventList.read(Path(self._outdir)/ f"{eventlist}")

        glon = self.gta.config['selection']['glon']
        glat = self.gta.config['selection']['glat']
        src_pos = SkyCoord(glon, glat, unit="deg", frame="galactic")
        counts = Map.create(skydir=src_pos, width=self.gta.config['binning']['roiwidth'],
            proj=self.gta.config['binning']['proj'], binsz=self.gta.config['binning']['binsz'],
            frame='galactic', axes=[self._energy_bins], dtype=float)
        counts.fill_by_coord({"skycoord": self._gammapy_events.radec, "energy": self._gammapy_events.energy})

        return counts

    def _load_fermi_irfs(self, counts, exposure = "bexpmap_00.fits", psf = "gtpsf_00.fits"):
        expmap = Map.read(Path(self._outdir) / f"{exposure}")
        axis = MapAxis.from_nodes(
            counts.geom.axes[0].center,
            name="energy_true",
            unit="MeV",
            interp="log"
        )

        irf = {}

        geom = WcsGeom(wcs=counts.geom.wcs, npix=counts.geom.npix, axes=[axis])
        exposure = expmap.interp_to_geom(geom)
        irf['exposure'] = exposure

        # PSF
        psf = PSFMap.read(Path(self._outdir) / f"{psf}", format="gtpsf")
        irf['psf'] = psf

        # Energy dispersion
        e_true = exposure.geom.axes["energy_true"]
        edisp = EDispMap.from_diagonal_response(energy_axis_true=e_true)
        irf['edisp'] = edisp

        return irf

    def _convert_model(self, fix_other_srcs=False):
        gammapy_models = []
        for src in self.gta.roi.sources:

            if src.name == "isodiff":
                gammapy_models.append(fermi_isotropic_diffuse_bkg(self.gta.config, src))
            elif src.name == "galdiff":
                gammapy_models.append(fermi_galactic_diffuse_bkg(self.gta.config, src))
            elif src.name == self.target_name:
                gammapy_models.append(fermipy2gammapy(self.gta.like, src))
            else:
                gammapy_models.append(fermipy2gammapy(self.gta.like, src, fix_pars = fix_other_srcs))

            self._logging.debug(f"A source model for {src.name} is converted.")

        return Models(gammapy_models)
